@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -40,22 +40,37 @@
 
 #ifdef USE_WIN32
 
-DISK_FILE *file_open(char *name, int wr) {
+DISK_FILE *file_open(char *name, FILE_MODE mode) {
     DISK_FILE *df;
-    LPTSTR tstr;
+    LPTSTR tname;
     HANDLE fh;
+    DWORD desired_access, creation_disposition;
 
     /* open file */
-    tstr=str2tstr(name);
-    fh=CreateFile(tstr, wr ? GENERIC_WRITE : GENERIC_READ,
-        FILE_SHARE_READ, NULL, wr ? OPEN_ALWAYS : OPEN_EXISTING,
-        FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL);
-    str_free(tstr);
-    if(fh==INVALID_HANDLE_VALUE) {
-        ioerror(name);
+    switch(mode) {
+    case FILE_MODE_READ:
+        desired_access=GENERIC_READ;
+        creation_disposition=OPEN_EXISTING;
+        break;
+    case FILE_MODE_APPEND:
+            /* reportedly more compatible than FILE_APPEND_DATA */
+        desired_access=GENERIC_WRITE;
+        creation_disposition=OPEN_ALWAYS; /* keep the data */
+        break;
+    case FILE_MODE_OVERWRITE:
+        desired_access=GENERIC_WRITE;
+        creation_disposition=CREATE_ALWAYS; /* remove the data */
+        break;
+    default: /* invalid mode */
         return NULL;
     }
-    if(wr) /* append */
+    tname=str2tstr(name);
+    fh=CreateFile(tname, desired_access, FILE_SHARE_READ, NULL,
+        creation_disposition, FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL);
+    str_free(tname); /* str_free() overwrites GetLastError() value */
+    if(fh==INVALID_HANDLE_VALUE)
+        return NULL;
+    if(mode==FILE_MODE_APPEND) /* workaround for FILE_APPEND_DATA */
         SetFilePointer(fh, 0, NULL, FILE_END);
 
     /* setup df structure */
@@ -74,15 +89,24 @@ DISK_FILE *file_fdopen(int fd) {
     return df;
 }
 
-DISK_FILE *file_open(char *name, int wr) {
+DISK_FILE *file_open(char *name, FILE_MODE mode) {
     DISK_FILE *df;
     int fd, flags;
 
     /* open file */
-    if(wr)
-        flags=O_CREAT|O_WRONLY|O_APPEND;
-    else
+    switch(mode) {
+    case FILE_MODE_READ:
         flags=O_RDONLY;
+        break;
+    case FILE_MODE_APPEND:
+        flags=O_CREAT|O_WRONLY|O_APPEND;
+        break;
+    case FILE_MODE_OVERWRITE:
+        flags=O_CREAT|O_WRONLY|O_TRUNC;
+        break;
+    default: /* invalid mode */
+        return NULL;
+    }
 #ifdef O_NONBLOCK
     flags|=O_NONBLOCK;
 #elif defined O_NDELAY
@@ -92,10 +116,8 @@ DISK_FILE *file_open(char *name, int wr) {
     flags|=O_CLOEXEC;
 #endif /* O_CLOEXEC */
     fd=open(name, flags, 0640);
-    if(fd<0) {
-        ioerror(name);
+    if(fd<0)
         return NULL;
-    }
 
     /* setup df structure */
     df=str_alloc(sizeof df);
@@ -178,18 +200,20 @@ int file_putline(DISK_FILE *df, char *line) {
 
 #ifdef USE_WIN32
 
-LPTSTR str2tstr(const LPSTR in) {
+LPTSTR str2tstr(LPCSTR in) {
     LPTSTR out;
     int len;
 
 #ifdef UNICODE
-    len=MultiByteToWideChar(CP_ACP, 0, in, -1, NULL, 0);
+    len=MultiByteToWideChar(CP_UTF8, 0, in, -1, NULL, 0);
     if(!len)
-        return NULL;
+        return str_tprintf(TEXT("MultiByteToWideChar() failed"));
     out=str_alloc((len+1)*sizeof(WCHAR));
-    len=MultiByteToWideChar(CP_ACP, 0, in, -1, out, len);
-    if(!len)
-        return NULL;
+    len=MultiByteToWideChar(CP_UTF8, 0, in, -1, out, len);
+    if(!len) {
+        str_free(out);
+        return str_tprintf(TEXT("MultiByteToWideChar() failed"));
+    }
 #else
     len=strlen(in);
     out=str_alloc(len+1);
@@ -198,18 +222,20 @@ LPTSTR str2tstr(const LPSTR in) {
     return out;
 }
 
-LPSTR tstr2str(const LPTSTR in) {
+LPSTR tstr2str(LPCTSTR in) {
     LPSTR out;
     int len;
 
 #ifdef UNICODE
-    len=WideCharToMultiByte(CP_ACP, 0, in, -1, NULL, 0, NULL, NULL);
+    len=WideCharToMultiByte(CP_UTF8, 0, in, -1, NULL, 0, NULL, NULL);
     if(!len)
-        return NULL;
+        return str_printf("WideCharToMultiByte() failed");
     out=str_alloc(len+1);
-    len=WideCharToMultiByte(CP_ACP, 0, in, -1, out, len, NULL, NULL);
-    if(!len)
-        return NULL;
+    len=WideCharToMultiByte(CP_UTF8, 0, in, -1, out, len, NULL, NULL);
+    if(!len) {
+        str_free(out);
+        return str_printf("WideCharToMultiByte() failed");
+    }
 #else
     len=strlen(in);
     out=str_alloc(len+1);
