@@ -54,7 +54,8 @@
 #else /* MSDN claims that _WIN32 is always defined */
 #define STUNNEL_PLATFORM "Win32"
 #endif
-#define SERVICE_NAME "stunnel"
+#define SERVICE_NAME TEXT("stunnel")
+#define SERVICE_DISPLAY_NAME TEXT("Stunnel SSL wrapper")
 #endif
 
 /* mingw-Patches-1825044 is missing in Debian Squeeze */
@@ -62,7 +63,7 @@ WINBASEAPI BOOL WINAPI CheckTokenMembership(HANDLE, PSID, PBOOL);
 
 /* prototypes */
 NOEXPORT BOOL CALLBACK enum_windows(HWND, LPARAM);
-NOEXPORT void parse_cmdline(LPSTR);
+NOEXPORT void gui_cmdline();
 NOEXPORT int initialize_winsock(void);
 NOEXPORT int gui_loop();
 
@@ -81,15 +82,15 @@ NOEXPORT void daemon_thread(void *);
 NOEXPORT void valid_config(void);
 NOEXPORT void invalid_config(void);
 NOEXPORT void update_peer_menu(void);
-NOEXPORT void update_tray_icon(const int);
-NOEXPORT void error_box(const LPSTR);
+NOEXPORT void update_tray(const int);
+NOEXPORT void error_box(LPCTSTR);
 NOEXPORT void edit_config(HWND);
 NOEXPORT BOOL is_admin(void);
 
 /* NT Service related function */
 #ifndef _WIN32_WCE
 NOEXPORT int service_initialize(void);
-NOEXPORT int service_install(LPTSTR);
+NOEXPORT int service_install(void);
 NOEXPORT int service_uninstall(void);
 NOEXPORT int service_start(void);
 NOEXPORT int service_stop(void);
@@ -97,6 +98,9 @@ NOEXPORT void WINAPI service_main(DWORD, LPTSTR *);
 NOEXPORT void WINAPI control_handler(DWORD);
 #endif /* !defined(_WIN32_WCE) */
 
+/* other functions */
+NOEXPORT LPTSTR get_params();
+
 /* global variables */
 static struct LIST {
   struct LIST *next;
@@ -127,7 +131,6 @@ static SERVICE_STATUS_HANDLE serviceStatusHandle=0;
 #endif
 
 static volatile int visible=0;
-static volatile int error_mode=1; /* no valid configuration was ever loaded */
 static HANDLE config_ready=NULL; /* reload without a valid configuration */
 static LONG new_logs=0;
 
@@ -148,25 +151,19 @@ int WINAPI WinMain(HINSTANCE this_instance, HINSTANCE prev_instance,
         LPSTR lpCmdLine,
 #endif
         int nCmdShow) {
-    LPSTR command_line;
-    char *c, stunnel_exe_path[MAX_PATH];
+    TCHAR stunnel_exe_path[MAX_PATH];
+    LPTSTR c;
 #ifndef _WIN32_WCE
-    char *errmsg;
+    LPTSTR errmsg;
 #endif
 
     (void)prev_instance; /* skip warning about unused parameter */
+    (void)lpCmdLine; /* skip warning about unused parameter */
     (void)nCmdShow; /* skip warning about unused parameter */
 
     str_init(); /* initialize per-thread string management */
     ghInst=this_instance;
-#ifdef _WIN32_WCE
-    command_line=tstr2str(lpCmdLine);
-#else
-    command_line=lpCmdLine;
-#endif
-
-    parse_cmdline(command_line); /* setup global cmdline structure */
-
+    gui_cmdline(); /* setup global cmdline structure */
     GetModuleFileName(0, stunnel_exe_path, MAX_PATH);
 
 #ifndef _WIN32_WCE
@@ -180,18 +177,19 @@ int WINAPI WinMain(HINSTANCE this_instance, HINSTANCE prev_instance,
 #endif
 
     /* set current working directory and engine path */
-    c=strrchr(stunnel_exe_path, '\\'); /* last backslash */
+    c=_tcsrchr(stunnel_exe_path, TEXT('\\')); /* last backslash */
     if(c) /* found */
-        c[1]='\0'; /* truncate program name */
+        c[1]=TEXT('\0'); /* truncate program name */
 #ifndef _WIN32_WCE
     if(!SetCurrentDirectory(stunnel_exe_path)) {
-        errmsg=str_printf("Cannot set directory to %s", stunnel_exe_path);
+        errmsg=str_tprintf(TEXT("Cannot set directory to %s"),
+            stunnel_exe_path);
         message_box(errmsg, MB_ICONERROR);
         str_free(errmsg);
         return 1;
     }
 #endif
-    _putenv_s("OPENSSL_ENGINES", stunnel_exe_path);
+    _tputenv_s(TEXT("OPENSSL_ENGINES"), stunnel_exe_path);
 
     if(initialize_winsock())
         return 1;
@@ -200,7 +198,7 @@ int WINAPI WinMain(HINSTANCE this_instance, HINSTANCE prev_instance,
     if(cmdline.service) /* "-service" must be processed before "-install" */
         return service_initialize();
     if(cmdline.install)
-        return service_install(command_line);
+        return service_install();
     if(cmdline.uninstall)
         return service_uninstall();
     if(cmdline.start)
@@ -216,9 +214,9 @@ int WINAPI WinMain(HINSTANCE this_instance, HINSTANCE prev_instance,
 NOEXPORT BOOL CALLBACK enum_windows(HWND other_window_handle, LPARAM lParam) {
     DWORD pid;
     HINSTANCE hInstance;
-    char window_exe_path[MAX_PATH];
     HANDLE process_handle;
-    char *stunnel_exe_path=(char *)lParam;
+    TCHAR window_exe_path[MAX_PATH];
+    LPTSTR stunnel_exe_path=(LPTSTR)lParam;
 
     if(!other_window_handle)
         return TRUE;
@@ -235,7 +233,7 @@ NOEXPORT BOOL CALLBACK enum_windows(HWND other_window_handle, LPARAM lParam) {
         CloseHandle(process_handle);
         return TRUE;
     }
-    if(strcmp(stunnel_exe_path, window_exe_path)) {
+    if(_tcscmp(stunnel_exe_path, window_exe_path)) {
         CloseHandle(process_handle);
         return TRUE;
     }
@@ -256,12 +254,11 @@ NOEXPORT BOOL CALLBACK enum_windows(HWND other_window_handle, LPARAM lParam) {
 
 #endif
 
-NOEXPORT void parse_cmdline(LPSTR command_line) {
+NOEXPORT void gui_cmdline() {
     char *line, *c, *opt;
 
-    line=str_dup(command_line);
+    line=tstr2str(get_params());
     memset(&cmdline, 0, sizeof cmdline);
-
     c=line;
     while(*c && (*c=='/' || *c=='-')) {
         opt=c;
@@ -283,12 +280,20 @@ NOEXPORT void parse_cmdline(LPSTR command_line) {
             cmdline.quiet=1;
         else if(!strcasecmp(opt+1, "exit"))
             cmdline.exit=1;
-        else { /* option to be processed in options.c */
+        else { /* an option to be processed in options.c */
             c=opt;
             break;
         }
     }
-    cmdline.config_file=*c ? str_dup(c) : NULL;
+    if(*c=='\"') { /* the option is within double quotes */
+        c++;
+        opt=c;
+        while(*c && *c!='\"') /* find the closing double quote */
+            c++;
+        *c='\0';
+    } else /* the option is simply the rest of the line */
+        opt=c;
+    cmdline.config_file=*opt ? str_dup(opt) : NULL;
     str_free(line);
 }
 
@@ -297,7 +302,7 @@ NOEXPORT int initialize_winsock() {
     static struct WSAData wsa_state;
 
     if(WSAStartup(MAKEWORD( 2, 2 ), &wsa_state)) {
-        message_box("Failed to initialize winsock", MB_ICONERROR);
+        message_box(TEXT("Failed to initialize winsock"), MB_ICONERROR);
         return 1; /* error */
     }
     resolver_init();
@@ -353,15 +358,17 @@ NOEXPORT int gui_loop() {
             EnableMenuItem(main_menu_handle, IDM_EDIT_CONFIG, MF_GRAYED);
             EnableMenuItem(main_menu_handle, IDM_SAVE_LOG, MF_GRAYED);
         }
-        if(tray_menu_handle) {
+        if(tray_menu_handle)
             EnableMenuItem(tray_menu_handle, IDM_EDIT_CONFIG, MF_GRAYED);
-        }
     }
 #endif
     /* auto-reset, non-signaled */
     config_ready=CreateEvent(NULL, FALSE, FALSE, NULL);
     _beginthread(daemon_thread, DEFAULT_STACK_SIZE, NULL);
 
+    /* setup periodic event to trigger update_logs() */
+    SetTimer(hwnd, 0x29a, 1000, NULL); /* single WM_TIMER event per second */
+
     while(GetMessage(&msg, NULL, 0, 0)) {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
@@ -388,11 +395,11 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
         /* create command bar */
         command_bar_handle=CommandBar_Create(ghInst, main_window_handle, 1);
         if(!command_bar_handle)
-            error_box("CommandBar_Create");
+            error_box(TEXT("CommandBar_Create"));
         if(!CommandBar_InsertMenubar(command_bar_handle, ghInst, IDM_MAINMENU, 0))
-            error_box("CommandBar_InsertMenubar");
+            error_box(TEXT("CommandBar_InsertMenubar"));
         if(!CommandBar_AddAdornments(command_bar_handle, 0, 0))
-            error_box("CommandBar_AddAdornments");
+            error_box(TEXT("CommandBar_AddAdornments"));
 #endif
 
         /* create child edit window */
@@ -425,8 +432,7 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
         return TRUE;
 
     case WM_TIMER:
-        if(visible)
-            update_logs();
+        update_logs();
         return TRUE;
 
     case WM_CLOSE:
@@ -438,19 +444,22 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
         if(tray_menu_handle)
             CheckMenuItem(tray_menu_handle, IDM_SHOW_LOG,
                 visible ? MF_CHECKED : MF_UNCHECKED);
-        if(visible)
-            update_logs();
+        update_logs();
         return TRUE;
 
     case WM_DESTROY:
 #ifdef _WIN32_WCE
         CommandBar_Destroy(command_bar_handle);
 #else
-        if(main_menu_handle)
+        if(main_menu_handle) {
             DestroyMenu(main_menu_handle);
+            main_menu_handle=NULL;
+        }
 #endif
-        if(tray_menu_handle)
+        if(tray_menu_handle) {
             DestroyMenu(tray_menu_handle);
+            tray_menu_handle=NULL;
+        }
         ZeroMemory(&nid, sizeof nid);
         nid.cbSize=sizeof nid;
         nid.hWnd=main_window_handle;
@@ -497,7 +506,7 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
             ShowWindow(main_window_handle, SW_HIDE); /* hide window */
             break;
         case IDM_EXIT:
-            if(!error_mode) /* signal_pipe is active */
+            if(num_clients>=0) /* signal_pipe is active */
                 signal_post(SIGNAL_TERMINATE);
             DestroyWindow(main_window_handle);
             break;
@@ -512,10 +521,10 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
 #endif
             break;
         case IDM_RELOAD_CONFIG:
-            if(error_mode) /* unlock daemon_thread */
-                SetEvent(config_ready);
-            else /* signal_pipe is active */
+            if(num_clients>=0) /* signal_pipe is active */
                 signal_post(SIGNAL_RELOAD_CONFIG);
+            else
+                SetEvent(config_ready); /* unlock daemon_thread() */
             break;
         case IDM_REOPEN_LOG:
             signal_post(SIGNAL_REOPEN_LOG);
@@ -588,7 +597,7 @@ NOEXPORT LRESULT CALLBACK window_proc(HWND main_window_handle,
         return TRUE;
 
     case WM_CLIENTS:
-        update_tray_icon((int)wParam);
+        update_tray((int)wParam);
         return TRUE;
     }
 
@@ -615,25 +624,25 @@ NOEXPORT LRESULT CALLBACK about_proc(HWND dialog_handle, UINT message,
 
 NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
         WPARAM wParam, LPARAM lParam) {
-    char *titlebar;
-    LPTSTR tstr;
+    LPTSTR titlebar;
     union {
         TCHAR txt[PEM_BUFSIZE];
         WORD len;
     } pass_dialog;
     WORD pass_len;
     char* pass_txt;
+    LPTSTR key_file_name;
 
     switch(message) {
     case WM_INITDIALOG:
         /* set the default push button to "Cancel" */
         SendMessage(dialog_handle, DM_SETDEFID, (WPARAM)IDCANCEL, (LPARAM)0);
 
-        titlebar=str_printf("Private key: %s", ui_data->section->key);
-        tstr=str2tstr(titlebar);
+        key_file_name=str2tstr(ui_data->section->key);
+        titlebar=str_tprintf(TEXT("Private key: %s"), key_file_name);
+        str_free(key_file_name);
+        SetWindowText(dialog_handle, titlebar);
         str_free(titlebar);
-        SetWindowText(dialog_handle, tstr);
-        str_free(tstr);
         return TRUE;
 
     case WM_COMMAND:
@@ -658,7 +667,7 @@ NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
                 (WPARAM)0 /* line 0 */, (LPARAM)pass_dialog.txt);
             pass_dialog.txt[pass_len]='\0'; /* null-terminate the string */
 
-            /* convert input password to ANSI string (as ui_data->pass) */
+            /* convert input password to UTF-8 string (as ui_data->pass) */
             pass_txt=tstr2str(pass_dialog.txt);
             strcpy(ui_data->pass, pass_txt);
             str_free(pass_txt);
@@ -727,7 +736,7 @@ NOEXPORT void save_log() {
     if(!GetSaveFileName(&ofn))
         return;
 
-    txt=log_txt(); /* need to convert the result to plain ASCII */
+    txt=log_txt(); /* need to convert the result to UTF-8 */
     str=tstr2str(txt);
     str_free(txt);
     save_text_file(file_name, str);
@@ -741,12 +750,12 @@ NOEXPORT int save_text_file(LPTSTR file_name, char *str) {
     file_handle=CreateFile(file_name, GENERIC_WRITE, 0, NULL,
         CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
     if(file_handle==INVALID_HANDLE_VALUE) {
-        error_box("CreateFile");
+        error_box(TEXT("CreateFile"));
         return 1;
     }
     if(!WriteFile(file_handle, str, strlen(str), &ignore, NULL)) {
         CloseHandle(file_handle);
-        error_box("WriteFile");
+        error_box(TEXT("WriteFile"));
         return 1;
     }
     CloseHandle(file_handle);
@@ -788,11 +797,11 @@ NOEXPORT void win_log(LPSTR line) {
 NOEXPORT void update_logs(void) {
     LPTSTR txt;
 
+    if(!visible)
+        return;
     if(!InterlockedExchange(&new_logs, 0))
         return;
     txt=log_txt();
-    if(!txt)
-        return;
     SetWindowText(edit_handle, txt);
     str_free(txt);
     SendMessage(edit_handle, WM_VSCROLL, (WPARAM)SB_BOTTOM, (LPARAM)0);
@@ -810,12 +819,11 @@ NOEXPORT LPTSTR log_txt(void) {
         memcpy(buff+ptr, curr->txt, curr->len*sizeof(TCHAR));
         ptr+=curr->len;
         if(curr->next) {
-            buff[ptr++]='\r';
-            buff[ptr++]='\n';
+            buff[ptr++]=TEXT('\r');
+            buff[ptr++]=TEXT('\n');
         }
     }
-    buff[ptr]='\0';
-
+    buff[ptr]=TEXT('\0');
     return buff;
 }
 
@@ -824,16 +832,17 @@ NOEXPORT LPTSTR log_txt(void) {
 NOEXPORT void daemon_thread(void *arg) {
     (void)arg; /* skip warning about unused parameter */
 
-    main_initialize();
+    main_init();
     /* get a valid configuration */
     while(main_configure(cmdline.config_file, NULL)) {
+        cmdline.config_file=NULL; /* don't retry commandline switches */
         unbind_ports(); /* in case initialization failed after bind_ports() */
         log_flush(LOG_MODE_ERROR); /* otherwise logs are buffered */
         PostMessage(hwnd, WM_INVALID_CONFIG, 0, 0); /* display error */
         WaitForSingleObject(config_ready, INFINITE);
         log_close(); /* prevent main_configure() from logging in error mode */
     }
-    error_mode=0; /* a valid configuration was loaded */
+    PostMessage(hwnd, WM_VALID_CONFIG, 0, 0);
 
     /* start the main loop */
     daemon_loop();
@@ -853,13 +862,14 @@ NOEXPORT void invalid_config() {
     ShowWindow(hwnd, SW_SHOWNORMAL); /* show window */
     SetForegroundWindow(hwnd); /* bring on top */
 
-    update_tray_icon(-1); /* error icon */
+    update_tray(-1); /* error icon */
+    update_peer_menu(); /* purge the list of sections */
 
     win_log("");
     s_log(LOG_ERR, "Server is down");
-    message_box("Stunnel server is down due to an error.\n"
-        "You need to exit and correct the problem.\n"
-        "Click OK to see the error log window.",
+    message_box(TEXT("Stunnel server is down due to an error.\n")
+        TEXT("You need to exit and correct the problem.\n")
+        TEXT("Click OK to see the error log window."),
         MB_ICONERROR);
 }
 
@@ -869,10 +879,8 @@ NOEXPORT void valid_config() {
         TEXT(STUNNEL_PLATFORM);
     SetWindowText(hwnd, win32_name);
 
-    if(global_options.option.taskbar) /* save menu resources */
-        update_tray_icon(num_clients); /* idle icon */
-
-    update_peer_menu();
+    update_tray(num_clients); /* idle or busy icon (on reload) */
+    update_peer_menu(); /* one menu item per section */
 
     /* enable IDM_REOPEN_LOG menu if a log file is used, disable otherwise */
 #ifndef _WIN32_WCE
@@ -890,9 +898,9 @@ NOEXPORT void update_peer_menu(void) {
     HMENU main_peer_list=NULL;
 #endif
     HMENU tray_peer_list=NULL;
-    char *str;
     unsigned int section_number;
     MENUITEMINFO mii;
+    LPTSTR servname;
 
     /* purge menu peer lists */
 #ifndef _WIN32_WCE
@@ -915,25 +923,22 @@ NOEXPORT void update_peer_menu(void) {
 
     section_number=0;
     for(section=service_options.next; section; section=section->next) {
-        /* setup section->file */
-        str=str_printf("peer-%s.pem", section->servname);
-        section->file=str2tstr(str);
-        str_free(str);
+        servname=str2tstr(section->servname);
 
         /* setup LPTSTR section->file */
-        str=str_printf("peer-%s.pem", section->servname);
-        section->file=str2tstr(str);
-        str_free(str);
+        section->file=str_tprintf(TEXT("peer-%s.pem"), servname);
 
-        /* setup (char *) section->help */
-        section->help=str_printf(
-            "Peer certificate chain has been saved.\n"
-            "Add the following lines to section [%s]:\n"
-            "\tCAfile = peer-%s.pem\n"
-            "\tverify = 3\n"
-            "to enable cryptographic authentication.\n"
-            "Then reload stunnel configuration file.",
-            section->servname, section->servname);
+        /* setup section->help */
+        section->help=str_tprintf(
+            TEXT("Peer certificate chain has been saved.\n")
+            TEXT("Add the following lines to section [%s]:\n")
+            TEXT("\tCAfile = peer-%s.pem\n")
+            TEXT("\tverify = 3\n")
+            TEXT("to enable cryptographic authentication.\n")
+            TEXT("Then reload stunnel configuration file."),
+            servname, servname);
+
+        str_free(servname);
 
         /* setup section->chain */
         section->chain=NULL;
@@ -961,23 +966,25 @@ NOEXPORT void update_peer_menu(void) {
 }
 
 ICON_IMAGE load_icon_default(ICON_TYPE type) {
-    WORD icon;
+    WORD idi;
+    ICON_IMAGE img;
 
     switch(type) {
     case ICON_ACTIVE:
-        icon=IDI_STUNNEL_ACTIVE;
+        idi=IDI_STUNNEL_ACTIVE;
         break;
     case ICON_ERROR:
-        icon=IDI_STUNNEL_ERROR;
+        idi=IDI_STUNNEL_ERROR;
         break;
     case ICON_IDLE:
-        icon=IDI_STUNNEL_IDLE;
+        idi=IDI_STUNNEL_IDLE;
         break;
     default:
         return NULL;
     }
-    return LoadImage(ghInst, MAKEINTRESOURCE(icon), IMAGE_ICON,
+    img=LoadImage(ghInst, MAKEINTRESOURCE(idi), IMAGE_ICON,
         GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), 0);
+    return DuplicateIcon(NULL, img);
 }
 
 ICON_IMAGE load_icon_file(const char *name) {
@@ -987,19 +994,33 @@ ICON_IMAGE load_icon_file(const char *name) {
     tname=str2tstr((LPSTR)name);
     icon=LoadImage(NULL, tname, IMAGE_ICON, GetSystemMetrics(SM_CXSMICON),
         GetSystemMetrics(SM_CYSMICON), LR_LOADFROMFILE);
+#ifndef _WIN32_WCE
+    icon=LoadImage(NULL, tname, IMAGE_ICON, GetSystemMetrics(SM_CXSMICON),
+        GetSystemMetrics(SM_CYSMICON), LR_LOADFROMFILE);
+#else
+    /* TODO: Implement a WCE version of LoadImage() */
+    /* icon=wceLoadIconFromFile(tname); */
+#endif
     str_free(tname);
     return icon;
 }
 
-NOEXPORT void update_tray_icon(const int num) {
+NOEXPORT void update_tray(const int num) {
     NOTIFYICONDATA nid;
     static ICON_TYPE previous_icon=ICON_NONE;
     ICON_TYPE current_icon;
+    LPTSTR tip;
 
-    if(!tray_menu_handle) { /* initialize taskbar */
-        tray_menu_handle=LoadMenu(ghInst, MAKEINTRESOURCE(IDM_TRAYMENU));
-        SetTimer(hwnd, 0x29a, 1000, NULL); /* 1-second timer */
+    if(!global_options.option.taskbar) {
+        /* release previously allocated menu resources */
+        if(tray_menu_handle) { /* disabled in the new configuration */
+            DestroyMenu(tray_menu_handle);
+            tray_menu_handle=NULL;
+        }
+        return;
     }
+    if(!tray_menu_handle) /* initialize taskbar */
+        tray_menu_handle=LoadMenu(ghInst, MAKEINTRESOURCE(IDM_TRAYMENU));
 
     ZeroMemory(&nid, sizeof nid);
     nid.cbSize=sizeof nid;
@@ -1008,15 +1029,18 @@ NOEXPORT void update_tray_icon(const int num) {
     nid.uCallbackMessage=WM_SYSTRAY; /* notification message */
     nid.hWnd=hwnd; /* window to receive notifications */
     if(num<0) {
-        _stprintf(nid.szTip, TEXT("Server is down"));
+        tip=str_tprintf(TEXT("Server is down"));
         current_icon=ICON_ERROR;
     } else if(num>0) {
-        _stprintf(nid.szTip, TEXT("%d active session(s)"), num);
+        tip=str_tprintf(TEXT("%d active session(s)"), num);
         current_icon=ICON_ACTIVE;
     } else {
-        _stprintf(nid.szTip, TEXT("Server is idle"));
+        tip=str_tprintf(TEXT("Server is idle"));
         current_icon=ICON_IDLE;
     }
+    _tcsncpy(nid.szTip, tip, 63);
+    nid.szTip[63]=TEXT('\0');
+    str_free(tip);
     nid.hIcon=global_options.icon[current_icon];
     if(current_icon!=previous_icon) {
         nid.uFlags|=NIF_ICON;
@@ -1029,31 +1053,24 @@ NOEXPORT void update_tray_icon(const int num) {
     Shell_NotifyIcon(NIM_ADD, &nid);
 }
 
-NOEXPORT void error_box(const LPSTR text) {
-    char *errmsg, *fullmsg;
-    LPTSTR tstr;
+NOEXPORT void error_box(LPCTSTR text) {
+    LPTSTR errmsg, fullmsg;
     long dw;
 
     dw=GetLastError();
     FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
         NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-        (LPTSTR)&tstr, 0, NULL);
-    errmsg=tstr2str(tstr);
-    LocalFree(tstr);
-    fullmsg=str_printf("%s: error %ld: %s", text, dw, errmsg);
-    str_free(errmsg);
+        (LPTSTR)&errmsg, 0, NULL);
+    fullmsg=str_tprintf(TEXT("%s: error %ld: %s"), text, dw, errmsg);
+    LocalFree(errmsg);
     message_box(fullmsg, MB_ICONERROR);
     str_free(fullmsg);
 }
 
-void message_box(const LPSTR text, const UINT type) {
-    LPTSTR tstr;
-
+void message_box(LPCTSTR text, const UINT type) {
     if(cmdline.quiet)
         return;
-    tstr=str2tstr(text);
-    MessageBox(hwnd, tstr, win32_name, type);
-    str_free(tstr);
+    MessageBox(hwnd, text, win32_name, type);
 }
 
 void ui_new_chain(const int section_number) {
@@ -1064,7 +1081,7 @@ void ui_new_log(const char *line) {
     SendMessage(hwnd, WM_LOG, (WPARAM)line, 0);
 }
 
-void ui_new_config(void) {
+void ui_config_reloaded(void) {
     PostMessage(hwnd, WM_VALID_CONFIG, 0, 0);
 }
 
@@ -1072,29 +1089,37 @@ void ui_clients(const int num) {
     PostMessage(hwnd, WM_CLIENTS, (WPARAM)num, 0);
 }
 
-NOEXPORT void edit_config(HWND main_window_handle) {
-    char cwd[MAX_PATH], *conf_path;
-
     /* TODO: port it to WCE */
+NOEXPORT void edit_config(HWND main_window_handle) {
+    TCHAR cwd[MAX_PATH];
+    LPTSTR conf_file, conf_path;
+
+    conf_file=str2tstr(configuration_file);
+    if(*conf_file==TEXT('\"')) {
+        conf_path=conf_file;
+    } else if(_tcschr(conf_file, TEXT('\\'))) {
+        conf_path=str_tprintf(TEXT("\"%s\""), conf_file);
+        str_free(conf_file);
+    } else {
+        GetCurrentDirectory(MAX_PATH, cwd);
+        conf_path=str_tprintf(TEXT("\"%s\\%s\""), cwd, conf_file);
+        str_free(conf_file);
+    }
+
     if(is_admin()) {
         ShellExecute(main_window_handle, TEXT("open"),
-            TEXT("notepad.exe"), configuration_file,
+            TEXT("notepad.exe"), conf_path,
             NULL, SW_SHOWNORMAL);
     } else { /* UAC workaround */
-        if(strchr(configuration_file, '\\')) {
-            conf_path=str_dup(configuration_file);
-        } else {
-            GetCurrentDirectory(MAX_PATH, cwd);
-            conf_path=str_printf("%s\\%s", cwd, configuration_file);
-        }
         ShellExecute(main_window_handle, TEXT("runas"),
             TEXT("notepad.exe"), conf_path,
             NULL, SW_SHOWNORMAL);
-        str_free(conf_path);
     }
+    str_free(conf_path);
 }
 
 NOEXPORT BOOL is_admin(void) {
+#ifndef _WIN32_WCE
     SID_IDENTIFIER_AUTHORITY NtAuthority={SECURITY_NT_AUTHORITY};
     PSID admin_group;
     BOOL retval;
@@ -1108,6 +1133,9 @@ NOEXPORT BOOL is_admin(void) {
         FreeSid(admin_group);
     }
     return retval;
+#else
+    return TRUE; /* always an admin on WCE */
+#endif
 }
 
 /**************************************** windows service */
@@ -1119,36 +1147,48 @@ NOEXPORT int service_initialize(void) {
 
     serviceTable[0].lpServiceName=SERVICE_NAME;
     serviceTable[0].lpServiceProc=service_main;
-    global_options.option.taskbar=0; /* disable taskbar for security */
+    /* disable taskbar for security */
+    /* global_options.option.taskbar=0; */
     if(!StartServiceCtrlDispatcher(serviceTable)) {
-        error_box("StartServiceCtrlDispatcher");
+        error_box(TEXT("StartServiceCtrlDispatcher"));
         return 1;
     }
     return 0; /* NT service started */
 }
 
-NOEXPORT int service_install(LPSTR command_line) {
+#define DESCR_LEN 256
+
+NOEXPORT int service_install() {
     SC_HANDLE scm, service;
-    char stunnel_exe_path[MAX_PATH], *service_path;
+    TCHAR stunnel_exe_path[MAX_PATH];
+    LPTSTR service_path;
+    TCHAR descr_str[DESCR_LEN];
+    SERVICE_DESCRIPTION descr;
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
     if(!scm) {
-        error_box("OpenSCManager");
+        error_box(TEXT("OpenSCManager"));
         return 1;
     }
     GetModuleFileName(0, stunnel_exe_path, MAX_PATH);
-    service_path=str_printf("\"%s\" -service %s", stunnel_exe_path, command_line);
-    service=CreateService(scm, SERVICE_NAME, SERVICE_NAME, SERVICE_ALL_ACCESS,
+    service_path=str_tprintf(TEXT("\"%s\" -service %s"),
+        stunnel_exe_path, get_params());
+    service=CreateService(scm, SERVICE_NAME, SERVICE_DISPLAY_NAME,
+        SERVICE_ALL_ACCESS,
         SERVICE_WIN32_OWN_PROCESS|SERVICE_INTERACTIVE_PROCESS,
         SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, service_path,
-        NULL, NULL, NULL, NULL, NULL);
+        NULL, NULL, TEXT("TCPIP\0"), NULL, NULL);
     str_free(service_path);
     if(!service) {
-        error_box("CreateService");
+        error_box(TEXT("CreateService"));
         CloseServiceHandle(scm);
         return 1;
     }
-    message_box("Service installed", MB_ICONINFORMATION);
+    if(LoadString(ghInst, IDS_SERVICE_DESC, descr_str, DESCR_LEN)) {
+        descr.lpDescription=descr_str;
+        ChangeServiceConfig2(service, SERVICE_CONFIG_DESCRIPTION, &descr);
+    }
+    message_box(TEXT("Service installed"), MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
@@ -1160,34 +1200,34 @@ NOEXPORT int service_uninstall(void) {
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
     if(!scm) {
-        error_box("OpenSCManager");
+        error_box(TEXT("OpenSCManager"));
         return 1;
     }
     service=OpenService(scm, SERVICE_NAME, SERVICE_QUERY_STATUS|DELETE);
     if(!service) {
-        error_box("OpenService");
+        error_box(TEXT("OpenService"));
         CloseServiceHandle(scm);
         return 1;
     }
     if(!QueryServiceStatus(service, &serviceStatus)) {
-        error_box("QueryServiceStatus");
+        error_box(TEXT("QueryServiceStatus"));
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
     if(serviceStatus.dwCurrentState!=SERVICE_STOPPED) {
-        message_box("The service is still running", MB_ICONERROR);
+        message_box(TEXT("The service is still running"), MB_ICONERROR);
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
     if(!DeleteService(service)) {
-        error_box("DeleteService");
+        error_box(TEXT("DeleteService"));
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
-    message_box("Service uninstalled", MB_ICONINFORMATION);
+    message_box(TEXT("Service uninstalled"), MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
@@ -1199,17 +1239,17 @@ NOEXPORT int service_start(void) {
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
     if(!scm) {
-        error_box("OpenSCManager");
+        error_box(TEXT("OpenSCManager"));
         return 1;
     }
     service=OpenService(scm, SERVICE_NAME, SERVICE_QUERY_STATUS|SERVICE_START);
     if(!service) {
-        error_box("OpenService");
+        error_box(TEXT("OpenService"));
         CloseServiceHandle(scm);
         return 1;
     }
     if(!StartService(service, 0, NULL)) {
-        error_box("StartService");
+        error_box(TEXT("StartService"));
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
@@ -1217,19 +1257,19 @@ NOEXPORT int service_start(void) {
     do {
         Sleep(1000);
         if(!QueryServiceStatus(service, &serviceStatus)) {
-            error_box("QueryServiceStatus");
+            error_box(TEXT("QueryServiceStatus"));
             CloseServiceHandle(service);
             CloseServiceHandle(scm);
             return 1;
         }
     } while(serviceStatus.dwCurrentState==SERVICE_START_PENDING);
     if(serviceStatus.dwCurrentState!=SERVICE_RUNNING) {
-        message_box("Failed to start service", MB_ICONERROR);
+        message_box(TEXT("Failed to start service"), MB_ICONERROR);
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
-    message_box("Service started", MB_ICONINFORMATION);
+    message_box(TEXT("Service started"), MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
@@ -1241,29 +1281,29 @@ NOEXPORT int service_stop(void) {
 
     scm=OpenSCManager(0, 0, SC_MANAGER_CONNECT);
     if(!scm) {
-        error_box("OpenSCManager");
+        error_box(TEXT("OpenSCManager"));
         return 1;
     }
     service=OpenService(scm, SERVICE_NAME, SERVICE_QUERY_STATUS|SERVICE_STOP);
     if(!service) {
-        error_box("OpenService");
+        error_box(TEXT("OpenService"));
         CloseServiceHandle(scm);
         return 1;
     }
     if(!QueryServiceStatus(service, &serviceStatus)) {
-        error_box("QueryServiceStatus");
+        error_box(TEXT("QueryServiceStatus"));
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
     if(serviceStatus.dwCurrentState==SERVICE_STOPPED) {
-        message_box("The service is already stopped", MB_ICONERROR);
+        message_box(TEXT("The service is already stopped"), MB_ICONERROR);
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
     }
     if(!ControlService(service, SERVICE_CONTROL_STOP, &serviceStatus)) {
-        error_box("ControlService");
+        error_box(TEXT("ControlService"));
         CloseServiceHandle(service);
         CloseServiceHandle(scm);
         return 1;
@@ -1271,13 +1311,13 @@ NOEXPORT int service_stop(void) {
     do {
         Sleep(1000);
         if(!QueryServiceStatus(service, &serviceStatus)) {
-            error_box("QueryServiceStatus");
+            error_box(TEXT("QueryServiceStatus"));
             CloseServiceHandle(service);
             CloseServiceHandle(scm);
             return 1;
         }
     } while(serviceStatus.dwCurrentState!=SERVICE_STOPPED);
-    message_box("Service stopped", MB_ICONINFORMATION);
+    message_box(TEXT("Service stopped"), MB_ICONINFORMATION);
     CloseServiceHandle(service);
     CloseServiceHandle(scm);
     return 0;
@@ -1354,4 +1394,27 @@ NOEXPORT void WINAPI control_handler(DWORD controlCode) {
 
 #endif /* !defined(_WIN32_WCE) */
 
+/**************************************** other functions */
+
+NOEXPORT LPTSTR get_params() {
+    LPTSTR c;
+    TCHAR s;
+
+    c=GetCommandLine();
+    if(*c==TEXT('\"')) {
+        s=TEXT('\"');
+        ++c;
+    } else {
+        s=TEXT(' ');
+    }
+    for(; *c; ++c)
+        if(*c==s) {
+            ++c;
+            break;
+        }
+    while(*c==TEXT(' '))
+        ++c;
+    return c;
+}
+
 /* end of ui_win_gui.c */
